#include "common.h"
#include "tdrocr.h"
#include "utils.h"
#include "landmark.h"

using namespace cv;
using namespace std;

//#define REMOVE_SIGNAL_LINE

int isNewCamera();
int isOldCamera();
int postprocess_card_ai(Mat& src, int camera, std::vector<BoxfWithLandmarks>&detected_boxes_kps, vector<Point> &corner, float scale);
int detect_card_ai(Mat& src, std::vector<BoxfWithLandmarks> &detected_boxes_kps);

int areaSize(Mat& img, RotatedRect& roi_rect, int last, float& areaRatio);

// 卡片粗定位外接矩形和旋转矩形
typedef struct _CardRect {
  vector<Rect> boundRect;
  vector<RotatedRect> rotatedRect;
} CardRect;

// 轮廓点数据结构，c为中心点，p1和p2为角边上的点
struct CORNER {
  Point p1;
  Point c;
  Point p2;
};

/*************************************************
Function:       highlightDet1
Description:    高光检测函数。
                之前方法先进行将灰度范围归一化到0~1，再进行二值化
        现在不再归一化，将之前的归一化阈值换算到0-255的区间上。可以提速。
Previous:       highlightDet
Input:          src: 待处理的图像
Output:         dst: 二值化后的图像
Return:         正确输出返回0
*************************************************/
int highlightDet1(Mat& src, Mat& dst) {
  threshold(src, dst, 247, 255, cv::THRESH_BINARY);
  return 0;
}

/*************************************************
Function:       SkinDetect7
Description:    使用YCrCb色系提取肤色阈值。
                原方法为SkinDetect，此处仅优化书写格式，可以提速。
Input:          src:      待提取肤色区域的BGR图像
Output:         skinMask: 肤色阈值图
Used:           detect_Card
return：		执行成功返回0；
            输入、输出图片像素不匹配返回-1；
            输出图片仅支持单通道，否则返回-2.
*************************************************/
int SkinDetect7(Mat& src, Mat& skinMask) {
  Mat ycbcrImg;
  cvtColor(src, ycbcrImg, CV_BGR2YCrCb);
  uchar* ptYCbCr = ycbcrImg.data;
  uchar* ptMask = skinMask.data;
  int rows = ycbcrImg.rows;
  int cols = ycbcrImg.cols;
  int ycbcr_step = ycbcrImg.step[0];
  int skinMask_step = skinMask.step[0];

  if (skinMask.rows != rows || skinMask.cols != cols) return -1;
  if (skinMask.channels() != 1) return -2;

  for (int i = 0; i < rows; ++i) {
    int offset_j = 0;
    for (int j = 0; j < cols; ++j) {
      // 每个像素元素包括3个通道
      int Cb = (int)ptYCbCr[offset_j + 2];
      int Cr = (int)ptYCbCr[offset_j + 1];
      if (77 <= Cb && Cb <= 130 && 130 <= Cr && Cr <= 173) {
        // 单通道图像，不需要乘以通道数
        *(ptMask + j) = 255;
      }
      offset_j += 3;
    }
    ptYCbCr += ycbcr_step;
    ptMask += skinMask_step;
  }

  return 0;
}

/*************************************************
Function:       corner_two
Description:    利用定位到两个相邻角点，补齐四个角点
Input:          card_gray: 灰度图像
                corner: 输入两个角点
                card_w: 卡片宽度
                card_h: 卡片高度
Output:         corner: 输出四个角点
*************************************************/
void corner_two(Mat& card_gray, vector<Point>& corner, int card_w, int card_h) {
  // int card_w = 210, card_h = 132;
  sort(corner.begin(), corner.end(), compoint_x);
  int sideLen = pointLen(corner[0], corner[1]);
  double theta =
      atan2(abs(corner[1].y - corner[0].y), corner[1].x - corner[0].x);
  double scalingFactor = (sideLen > (card_w + card_h) / 2) ? card_h : card_w;
  int diff_x = scalingFactor * sin(theta);
  int diff_y = scalingFactor * cos(theta);

  /*向四个方向补角*/
  Point tmp1, tmp2;
  tmp1.x = corner[0].x - diff_x;
  tmp2.x = corner[1].x - diff_x;
  // 根据当前两个角的位置情况进行判断
  if (corner[0].y >= corner[1].y) {  // 右下
    tmp1.y = corner[0].y - diff_y;
    tmp2.y = corner[1].y - diff_y;
    if ((0 - tmp1.y) > 20 || (0 - tmp2.y) > 20) {
      tmp1.y = corner[0].y + diff_y;
      tmp2.y = corner[1].y + diff_y;
      tmp1.x = corner[0].x + diff_x;
      tmp2.x = corner[1].x + diff_x;
    }
  } else {  // 右上
    tmp1.y = corner[0].y + diff_y;
    tmp2.y = corner[1].y + diff_y;
    if ((tmp1.y - card_gray.rows) > 20 || (tmp2.y - card_gray.rows) > 20) {
      tmp1.y = corner[0].y - diff_y;
      tmp2.y = corner[1].y - diff_y;
      tmp1.x = corner[0].x + diff_x;
      tmp2.x = corner[1].x + diff_x;
    }
  }

  if ((0 - tmp1.x) > 20 || (0 - tmp2.x) > 20) {  // 左上
    tmp1.x = corner[0].x + diff_x;
    tmp2.x = corner[1].x + diff_x;
    if (corner[0].y >= corner[1].y) {
      tmp1.y = corner[0].y + diff_y;
      tmp2.y = corner[1].y + diff_y;
      if ((tmp1.y - card_gray.rows) > 20 || (tmp2.y - card_gray.rows) > 20) {
        tmp1.y = corner[0].y - diff_y;
        tmp2.y = corner[1].y - diff_y;
        tmp1.x = corner[0].x - diff_x;
        tmp2.x = corner[1].x - diff_x;
      }
    } else {  // 左下
      tmp1.y = corner[0].y - diff_y;
      tmp2.y = corner[1].y - diff_y;
      if ((0 - tmp1.y) > 20 || (0 - tmp2.y) > 20) {
        tmp1.y = corner[0].y + diff_y;
        tmp2.y = corner[1].y + diff_y;
        tmp1.x = corner[0].x - diff_x;
        tmp2.x = corner[1].x - diff_x;
      }
    }
  }
  if ((0 - tmp1.x) < 20 && (tmp1.x - card_gray.cols) < 20 &&
      (0 - tmp1.y) < 20 && (tmp1.y - card_gray.rows) < 20) {
    corner.push_back(tmp1);
  }
  if ((0 - tmp2.x) < 20 && (tmp2.x - card_gray.cols) < 20 &&
      (0 - tmp2.y) < 20 && (tmp2.y - card_gray.rows) < 20) {
    corner.push_back(tmp2);
  }
}

/*************************************************
Function:       cornertwo_Select
Description:    候选角点取两两组合，输出补齐后白点占比最大的四个点
Calls:          vertical2: 判断两条由两个点连线形成的直线是否垂直
                CrossPoint: 计算两条直线的交点
                CircleCross:
计算以两个对角线点为圆心，卡片长或宽为半径的圆的交点 corner_two:
由两个相邻角点补齐四个点 areaSize: 计算轮廓内白色占比

Input:          src: 输入图像
                corner: 所有候选角点
                corners: 候选角点和边
                reaRatio: 补齐后白点占比
                last:是否为第二次计算标识符，如果第二次计算，白点占比用不同方式计算
                card_w: 卡片宽度
                card_h: 卡片高度
Output:         corner: 输出最终补齐的四个角点
Return:         正确输出四个角点返回0， 否则返回-1
*************************************************/
int cornertwo_Select(Mat& src, vector<Point>& corner, vector<CORNER>& corners,
                     float& areaRatio, int last, int card_w, int card_h) {
  int ret = -1;
  int djerror = 20;  // 对角线长度误差

  vector<Point> cornerVer;  // = corner;
  // TODO 修改点1: 将仅需要计算一次并在循环中多次调用的乘法操作提到循环外。
  int carddj = sqrt(pow(card_w, 2) + pow(card_h, 2));  // 卡片对角线长度
  int lenError = 15;                                   // 对角线长度误差

  for (int i = 0; i < corner.size(); i++) {
    Point tmp;
    tmp.x = max(0, corner[i].x);
    tmp.x = min(tmp.x, src.cols - 1);
    tmp.y = max(0, corner[i].y);
    tmp.y = min(tmp.y, src.rows - 1);
    cornerVer.push_back(tmp);
  }  // 卡片候选点，需保证在图片范围内
  float areaRatioMax = 0.0;

  for (int i = 0; i < cornerVer.size() - 1; ++i) {
    for (int j = i + 1; j < cornerVer.size(); ++j) {
      float areaRatioTmp = 0;
      vector<Point> cornerTmp;
      vector<CORNER> acornerTmp;
      Point tmp1 = cornerVer[i];
      Point tmp2 = cornerVer[j];
      cornerTmp.push_back(tmp1);
      cornerTmp.push_back(tmp2);
      if (i >= corners.size() || j >= corners.size()) continue;

      acornerTmp.push_back(corners[i]);  // 卡片候选角点
      acornerTmp.push_back(corners[j]);

      int p2p_len = pointLen(cornerTmp[0], cornerTmp[1]);

      if (abs(p2p_len - carddj) < lenError)  // 候选点满足对角线长度
      {
        Point a_1, b_1, c_1, d_1, a_2, b_2, c_2, d_2;
        a_1 = acornerTmp[0].c;
        b_1 = acornerTmp[0].p1;
        c_1 = acornerTmp[1].c;
        d_1 = acornerTmp[1].p2;

        a_2 = acornerTmp[0].c;
        b_2 = acornerTmp[0].p2;
        c_2 = acornerTmp[1].c;
        d_2 = acornerTmp[1].p1;

        double dj1 = calc_dist(c_1, d_1, a_1);
        double dj2 = calc_dist(c_2, d_2, a_2);
        if (vertical2(a_1, b_1, c_1, d_1) && vertical2(a_2, b_2, c_2, d_2) &&
            ((abs(dj1 - card_w) < djerror && abs(dj2 - card_h) < djerror) ||
             (abs(dj1 - card_h) < djerror && abs(dj2 - card_w) < djerror)))
        // 对角点满足两边互相垂直，且到其对角点的距离满足卡片的长或宽
        {
          Point cro1, cro2;
          Point tmp_3 = CrossPoint(a_1, b_1, c_1, d_1);
          Point tmp_4 =
              CrossPoint(a_2, b_2, c_2, d_2);  // 计算对角点边的交点为第一候选点
          Point tmp_5 =
              CircleCross(a_1, c_1, tmp_3, tmp_4, cro1, card_w, card_h);
          Point tmp_6 = CircleCross(
              c_2, a_2, tmp_3, tmp_4, cro2, card_w,
              card_h);  // 计算以对角点为圆心，卡片长宽为半径圆的交点为第二候选点

          /*选择第一和第二候选点与对角点组成的角更接近90度的作为最终补全的点*/
          double anglel1 = lineAngle(a_1, c_1, tmp_5);
          double anglel2 = lineAngle(a_1, c_1, cro1);

          Point t1 =
              abs(abs(anglel1) - 90) < abs(abs(anglel2) - 90) ? tmp_5 : cro1;
          anglel1 = lineAngle(a_1, c_1, tmp_6);
          anglel2 = lineAngle(a_1, c_1, cro2);

          Point t2 =
              abs(abs(anglel1) - 90) < abs(abs(anglel2) - 90) ? tmp_6 : cro2;
          cornerTmp.push_back(t1);  // 0805
          cornerTmp.push_back(t2);
        }
      }
      /*满足条件的边的点，补齐另外两点*/
      if ((abs(p2p_len - card_w) < lenError) ||
          (abs(p2p_len - card_h) < lenError)) {
        corner_two(src, cornerTmp, card_w, card_h);
      }

      if (cornerTmp.size() < 4) continue;

      /*保证找到的角点位置不越界*/
      for (int i = 0; i < cornerTmp.size(); i++) {
        cornerTmp[i].x = max(0, cornerTmp[i].x);
        cornerTmp[i].x = min(cornerTmp[i].x, src.cols - 1);
        cornerTmp[i].y = max(0, cornerTmp[i].y);
        cornerTmp[i].y = min(cornerTmp[i].y, src.rows - 1);
      }

      /*轮廓面积与card面积比值最大的轮廓 作为最终补全的点*/
      RotatedRect roiTmp = minAreaRect(cornerTmp);
      ret = areaSize(src, roiTmp, last, areaRatioTmp);
      if (ret == 0 && areaRatioTmp > areaRatioMax) {  // + 0.001) {
        areaRatioMax = areaRatioTmp;
        areaRatio = areaRatioMax;
        corner.assign(cornerTmp.begin(), cornerTmp.end());
      }
    }
  }

  return ret;
}

/*************************************************
Function:       detectCorner
Description:
对卡片分割后，确定轮廓上满足卡片条件的候选垂直点，补齐后计算roi区域占比，得到最终的卡片定位点
Calls:          vertical2_wx: 判断两条由两个点连线形成的直线是否垂直
                vertical3_wx: 判断轮廓点是否为垂直点
                cornerDirect: 判断角的方向是否满足条件
                calc_dist: 计算点到直线的距离
                CrossPoint: 计算两条直线的交点
                vertical_dj: 判断满足对角条件的点边是否垂直
                cornertwo_Select:
候选角点取两两组合，输出补齐后白点占比最大的四个点 cornerOvertwo_Select:
候选角点取三、四组合，输出补齐后白点占比最大的四个点

Input:          card_gray: 卡片灰度图像
                card_threshold: 卡片二值图
                contours: 卡片分割轮廓
                areaRatio: 补齐后白点占比
                last:
是否为第二次计算标识符，如果第二次计算，白点占比用不同方式计算 card_w: 卡片宽度
                card_h: 卡片高度
Output:         corner: 输出最终补齐的四个角点
Return:         成功检测四个角点返回0， 否则返回-1
*************************************************/
int detectCorner(Mat& card_gray, Mat& card_threshold,
                 vector<vector<Point>>& contours, vector<Point>& corner,
                 float& areaRatio, int last, int card_w, int card_h) {
  vector<CORNER> corners;       // 满足条件点
  vector<CORNER> corners_cand;  // 直角点候选点,点为倒角两边交点
  vector<Point> cornerstmp;  // 直角点候选点，点为倒角的中点，仅计算长度

  if (contours.size() < 1) return -1;

  int lenDiff = 20;  // 长度计算误差
  /*
  vector<Moments> mu(contours.size());
  vector<Point2f> mc(contours.size());
  */
  // TODO 修改点1:
  // 结构优化，筛选候选点时用临时变量替代数组变量。代码看着清晰且提速。
  // 候选点间的首尾关系统一用数组下标取余的方式处理，看着清晰。
  // 第一波筛角，筛选出直角和满足条件的钝角【卡片四端尖磨到或剪到的那种，几乎还是直角】
  // 判断轮廓上的点是否为垂直点
  for (int i = 0; i < contours.size(); i++) {
    int contour_item_pt_size = contours[i].size();
    if (contour_item_pt_size < 3) {
      continue;
    }
    for (int j = 0; j < contour_item_pt_size; j++) {
      Point cur_pt, last_pt, cur_next_1_pt, cur_next_2_pt;

      cur_pt = contours[i][j];
      // 取模处理轮廓点下标是头尾点的情况
      last_pt =
          contours[i][(j - 1 + contour_item_pt_size) % contour_item_pt_size];
      cur_next_1_pt = contours[i][(j + 1) % contour_item_pt_size];
      cur_next_2_pt = contours[i][(j + 2) % contour_item_pt_size];
      bool ver = false;
      bool ver1 = false;
      CORNER cr;

      // 判断当前点和两边组成的角是否为直角
      ver1 = vertical3_wx(last_pt, cur_pt, cur_next_1_pt);
      // 非直角判断是否为倒角情况
      if (pointLen(cur_pt, cur_next_1_pt) < lenDiff && !ver1) {
        // 判断两条由两点构成向量，是否正交
        ver = vertical2_wx(last_pt, cur_pt, cur_next_1_pt, cur_next_2_pt);
        if (ver) {
          // 若正交，则求这俩向量所在直线的交点
          Point newc =
              CrossPoint(last_pt, cur_pt, cur_next_1_pt, cur_next_2_pt);
          Point newmc;
          newmc.x = (cur_pt.x + cur_next_1_pt.x) / 2;
          newmc.y = (cur_pt.y + cur_next_1_pt.y) / 2;
          cr.p1 = last_pt;
          cr.c = newc;
          cr.p2 = cur_next_2_pt;
          // 候选点为倒角两边的交点
          corners_cand.push_back(cr);
          // 候选点为倒角两端点的中点，近计算距离用
          // TODO:存的是谁
          cornerstmp.push_back(newmc);
        }
      }
      // 非倒角垂直点
      if (ver1) {
        cr.p1 = last_pt;
        cr.c = cur_pt;
        cr.p2 = cur_next_1_pt;
        corners_cand.push_back(cr);
        cornerstmp.push_back(cur_pt);
      }
    }
    /*
    mu[i] = moments(contours[i], false);
    mc[i] = Point2f(mu[i].m10 / mu[i].m00, mu[i].m01 / mu[i].m00);
    //绘制轮廓
    Mat drawing = Mat::zeros(card_threshold.size(), CV_8UC3);
    RNG rng(12345);
    vector<Vec4i>hierarchy;
    for (int i = 0; i < contours.size(); i++)
    {
        Scalar color = Scalar(rng.uniform(0, 255), rng.uniform(0, 255),
    rng.uniform(0, 255)); drawContours(drawing, contours, i, color, 2, 8,
    hierarchy, 0, Point()); circle(drawing, mc[i], 4, color, -1, 8, 0);
    }

    namedWindow("轮廓图", WINDOW_AUTOSIZE);
    imshow("轮廓图", drawing);
    waitKey(0);
    */
  }

  int carddj = sqrt(pow(card_w, 2) + pow(card_h, 2));  // 卡片对角线长度
  int lenError = 10;                                   // 长度误差
  if (card_w > 220 || card_h > 140) lenError = 15;  // 旧摄像头长度误差
  int lineError = 10;                               // 对角线长度误差
  int cornerLen;
  bool ver, ver1;  //, ver2, ver3;
  if (corners_cand.size() < 2) return -1;

  // 第二波筛角，筛选出满足距离、方向的角
  for (int i = 0; i < corners_cand.size() - 1; i++) {
    for (int j = i + 1; j < corners_cand.size(); j++) {
      // 计算候选点之间的距离
      cornerLen = pointLen(cornerstmp[i], cornerstmp[j]);
      /*候选点长度满足卡片的长或宽*/
      if (abs(cornerLen - card_w) < lenError ||
          abs(cornerLen - card_h) < lenError) {
        ver = vertical2_wx(corners_cand[i].c, corners_cand[i].p1,
                           corners_cand[j].c, corners_cand[j].p1);

        // 候选点边上的点到另一角点边的距离
        double d1 = calc_dist(corners_cand[i].c, corners_cand[i].p1,
                              corners_cand[j].p2);
        double d2 = calc_dist(corners_cand[i].c, corners_cand[i].p2,
                              corners_cand[j].p1);
        // 判断角的方向是否满足条件
        int r1 = cornerDirect(corners_cand[i].c, corners_cand[i].p2,
                              corners_cand[j].c, corners_cand[j].p1);
        int r2 = cornerDirect(corners_cand[i].c, corners_cand[i].p1,
                              corners_cand[j].c, corners_cand[j].p2);

        // 两角边点的距离
        int dp1 = pointLen(corners_cand[i].p2, corners_cand[j].p1);
        int dp2 = pointLen(corners_cand[i].p1, corners_cand[j].p2);
        // 两角中心点的距离
        int dc = pointLen(corners_cand[i].c, corners_cand[j].c);

        if ((d1 < lineError && r1 > 0 && r2 < 0 && dc >= dp2 && ver) ||
            (d2 < lineError && r1 < 0 && r2 > 0 && dc >= dp1 && ver))
        // 角点距离满足边的距离，两边重合，且方向满足条件
        {
          corners.push_back(corners_cand[i]);
          corners.push_back(corners_cand[j]);
        }
      }

      /*候选点长度满足卡片对角线的长度*/
      else if (abs(cornerLen - carddj) < lenError) {
        int djerror = 20;  // 对角线长度误差
        // 对角点两边互相垂直
        ver = vertical_dj(corners_cand[i].p1, corners_cand[i].c,
                          corners_cand[j].c, corners_cand[j].p2);
        ver1 = vertical_dj(corners_cand[i].c, corners_cand[i].p2,
                           corners_cand[j].p1, corners_cand[j].c);
        // 点到线的距离
        double dj1 =
            calc_dist(corners_cand[j].c, corners_cand[j].p1, corners_cand[i].c);
        double dj2 = calc_dist(corners_cand[j].c, corners_cand[j].p2,
                               corners_cand[i].c);  // 角点到另一个点边的距离

        // 角点距离满足对角线长度，两边互相垂直，角点到另一个点边的距离满足卡片的长或宽
        if (ver1 && ver &&
            ((abs(dj1 - card_w) < djerror && abs(dj2 - card_h) < djerror) ||
             (abs(dj1 - card_h) < djerror && abs(dj2 - card_w) < djerror))) {
          int r1 = cornerDirect(corners_cand[i].c, corners_cand[i].p1,
                                corners_cand[j].c, corners_cand[j].p1);
          int r2 = cornerDirect(corners_cand[i].c, corners_cand[i].p2,
                                corners_cand[j].c, corners_cand[j].p2);
          // 角的方向满足条件
          if (r1 < 0 && r2 < 0) {
            corners.push_back(corners_cand[i]);
            corners.push_back(corners_cand[j]);
          }
        }
      }
    }
  }

  // vector<Point> cornercs;
  vector<CORNER> cornersq;
  map<pair<int, int>, int> cornercount;
  // 候选角去重
  // 通过角cornersq判定是否相同，但corner中仍会存在相同点的情况
  for (int i = 0; i < corners.size(); i++) {
    pair<int, int> pc;
    pc.first = corners[i].c.x;
    pc.second = corners[i].c.y;
    cornercount[pc] += 1;
    if (cornercount[pc] == 1) cornersq.push_back(corners[i]);
  }

  for (int i = 0; i < cornersq.size(); i++) {
    corner.push_back(cornersq[i].c);
  }
  if (corner.size() <= 1) return -1;
  cornertwo_Select(
      card_threshold, corner, cornersq, areaRatio, last, card_w,
      card_h);  // 候选角点取两两组合，输出补齐后白点占比最大的四个点
  if (areaRatio < 0.8) return -1;
  if (corner.size() < 4) return -1;

  return 0;
}

/*************************************************
Function:       cornerSearch
Description:    首先对输入图像进行图像预处理，
                之后寻找卡片轮廓，再进行多边形拟合，获取符合条件的卡片角点
Calls:          detectCorner：确定候选点和满足条件的最终角点
Input:          card_gray: 卡片灰度图
                card_threshold: 卡片二值图
                card_w: 卡片宽度
                card_h: 卡片高度
Output:         corner: 最终补齐的四个角点
                areaRatio: roi区域占比
Return:         正确输出四个角点返回0， 否则返回-1
*************************************************/
int cornerSearch(Mat& card_gray, Mat& card_threshold, vector<Point>& corner,
                 float& areaRatio, int card_w, int card_h) {
  int ret = 0;
  vector<vector<Point>> card_contours;
  vector<Vec4i> card_hierarchy;
  Mat element;
  // TODO 修改点1：将函数外图像处理，移到函数内
  // 中值→开运算→闭运算，可以去掉孤立的白色噪声、黑色噪声，并连通邻近的白色区域
  medianBlur(card_threshold, card_threshold, 3);
  element = getStructuringElement(MORPH_RECT, Size(5, 5));
  morphologyEx(card_threshold, card_threshold, CV_MOP_OPEN, element);
  element = getStructuringElement(MORPH_CROSS, Size(1, 5));  // 1, 3
  morphologyEx(card_threshold, card_threshold, CV_MOP_CLOSE, element);

  // 查找卡片轮廓
  findContours(card_threshold, card_contours, card_hierarchy, CV_RETR_EXTERNAL,
               CV_CHAIN_APPROX_SIMPLE);

  ret = -1;
  if (ret == -1) {
    corner.clear();
    vector<vector<Point>> card_contours_poly2;

    for (size_t i = 0; i < card_contours.size(); i++) {
      vector<Point> contours_tmp2;
      // 多边形拟合，轮廓点最小距离为3
      approxPolyDP(Mat(card_contours[i]), contours_tmp2, 3, true);
      RotatedRect rotatedtmp2 = minAreaRect(contours_tmp2);
      int w_tmp2 = max(rotatedtmp2.size.width, rotatedtmp2.size.height);
      int h_tmp2 = min(rotatedtmp2.size.width, rotatedtmp2.size.height);
      /*对于轮廓点之间距离小于15的点进行合并，取起始和结束的点，保证卡片倒角中只有前后两个点*/
      if (w_tmp2 > (card_w - 30) && h_tmp2 > (card_h - 30)) {
        vector<Point> tmpv;
        if (pointLen(contours_tmp2[contours_tmp2.size() - 1],
                     contours_tmp2[0]) > 15) {
          tmpv.push_back(contours_tmp2[0]);
        }
        for (int i = 1; i < contours_tmp2.size(); i++) {
          if (pointLen(contours_tmp2[i], contours_tmp2[i - 1]) > 15) {
            tmpv.push_back(contours_tmp2[i - 1]);
            tmpv.push_back(contours_tmp2[i]);
          }
        }
        if (pointLen(contours_tmp2[contours_tmp2.size() - 1],
                     contours_tmp2[0]) > 15) {
          tmpv.push_back(contours_tmp2[contours_tmp2.size() - 1]);
        }

        /*轮廓点去重*/
        vector<Point> tmpvqc;
        map<pair<int, int>, int> contourcount;
        for (int i = 0; i < tmpv.size(); i++) {
          pair<int, int> pc;
          pc.first = tmpv[i].x;
          pc.second = tmpv[i].y;
          contourcount[pc] += 1;
          if (contourcount[pc] == 1) tmpvqc.push_back(tmpv[i]);
        }

        card_contours_poly2.push_back(tmpvqc);
      }
    }
    /*轮廓点最小距离为3，计算白点占比用两种方法*/
    ret = detectCorner(card_gray, card_threshold, card_contours_poly2, corner,
                       areaRatio, 0, card_w, card_h);
    if (ret == -1)
      // TODO: 不添加？corner.clear();
      ret = detectCorner(card_gray, card_threshold, card_contours_poly2, corner,
                         areaRatio, 1, card_w, card_h);
    /*如果轮廓点最小距离为3未定位到卡片，取最小距离为10*/
    if (ret == -1) {
      corner.clear();
      vector<vector<Point>> card_contours_poly;
      for (size_t i = 0; i < card_contours.size(); i++) {
        vector<Point> contours_tmp;
        approxPolyDP(Mat(card_contours[i]), contours_tmp, 10, true);
        RotatedRect rotatedtmp = minAreaRect(contours_tmp);
        int w_tmp = max(rotatedtmp.size.width, rotatedtmp.size.height);
        int h_tmp = min(rotatedtmp.size.width, rotatedtmp.size.height);
        if (w_tmp > (card_w - 30) && h_tmp > (card_h - 30)) {
          card_contours_poly.push_back(contours_tmp);
        }
      }
      /*两种方法计算白点占比*/
      ret = detectCorner(card_gray, card_threshold, card_contours_poly, corner,
                         areaRatio, 0, card_w, card_h);
      if (ret == -1)
        ret = detectCorner(card_gray, card_threshold, card_contours_poly,
                           corner, areaRatio, 1, card_w, card_h);
    }

    if (ret == -1) return -1;
  }
  return ret;
}

/*************************************************
Function:       areaSize
Description:    计算卡片四个角点内区域与card面积的比值
                相比优化前省掉多余的操作
比如“获取旋转矩形不越界的外接矩形”，填充轮廓等
Input:          img: 卡片二值图像
                roi_rect: 定位到的卡片旋转矩形
                last: 是否为第二次计算标识
Output:         areaRatio: 白点占比
Return:         能够正常计算白点占比返回0， 否则返回-1
*************************************************/
int areaSize(Mat& img, RotatedRect& roi_rect, int last, float& areaRatio) {
  int ret = 0;
  areaRatio = 0.0;
  Rect_<float> safeBoundRct;
  bool isFormRect = calcSafeRect(roi_rect, img,
                                 safeBoundRct);  // 获取旋转矩形不越界的外接矩形
  if (!isFormRect) return -1;
  Mat cardVer;
  ret = imgRotate_2(img, cardVer, safeBoundRct, roi_rect,
                    1);  // 根据旋转矩形的角度旋转图像
  if (cardVer.cols <= 0 || cardVer.rows <= 0) return -1;

  if (ret == -1) return -1;
  Mat element;
  /*第一次计算闭运算核为5x5,第二次计算核为20x20*/
  if (last == 0) {
    element = getStructuringElement(MORPH_RECT, Size(5, 5));
  } else {
    element = getStructuringElement(MORPH_RECT, Size(20, 20));
  }
  morphologyEx(cardVer, cardVer, CV_MOP_CLOSE, element);
  double areaMax = 0;
  vector<vector<Point>> card_contours;
  vector<Vec4i> card_hierarchy;
  findContours(cardVer, card_contours, card_hierarchy, CV_RETR_EXTERNAL,
               CV_CHAIN_APPROX_SIMPLE);
  sort(card_contours.begin(), card_contours.end(),
       compContours);  // 查找轮廓，按照大小排序
  if (card_contours.size() < 1) return ret;
  Mat cardVer1 = Mat::zeros(cardVer.size(), CV_8UC1);
  drawContours(cardVer1, card_contours, 0, Scalar(255),
               CV_FILLED);  // 最大轮廓填充
  findContours(cardVer1, card_contours, card_hierarchy, CV_RETR_EXTERNAL,
               CV_CHAIN_APPROX_SIMPLE);  // 填充后再获取轮廓
  vector<vector<Point>> card_contours_poly;
  for (size_t i = 0; i < card_contours.size(); i++) {
    vector<Point> contours_tmp;
    approxPolyDP(Mat(card_contours[i]), contours_tmp, 3, true);
    double areaTmp = contourArea(contours_tmp, false);
    if (areaTmp > areaMax) areaMax = areaTmp;
    card_contours_poly.push_back(contours_tmp);
  }
  areaRatio = areaMax / (cardVer.cols * cardVer.rows *
                         1.0);  // 取面积最大的轮廓计算白点占比

  return ret;
}
/*************************************************
Function:       cardCorner2
Description:    应用自适应二值化61,0分割图片，去除肤色，定位角点
Calls:          cornerSearch：卡片轮廓多边形拟合，获取符合条件的卡片角点
Input:          card_gray: 卡片灰度图
                skinMask: 肤色掩膜
                card_w: 卡片宽度
                ard_h: 卡片高度
Output:         corner: 定位到的角点
                areaRatio: 白点占比
Return:         正确提取四个角点返回0，否则返回-1
*************************************************/
int cardCorner2(Mat& card_gray, Mat& skinMask, vector<Point>& corner,
                float& areaRatio, int card_w, int card_h) {
  corner.clear();
  int ret = 0;
  Mat card_threshold1, card_threshold2;
  // 自适应二值化
  adaptiveThreshold(card_gray, card_threshold1, 255, CV_ADAPTIVE_THRESH_MEAN_C,
                    CV_THRESH_BINARY, 61, 0);
  card_threshold2 = card_threshold1 - skinMask;
  // TODO
  // 修改点：调用cornerSearch方法之前仅做功能相关操作，将图像优化相关操作放在cornerSearch中统一完成
  ret = cornerSearch(card_gray, card_threshold2, corner, areaRatio, card_w,
                     card_h);  // 卡片定位
  if (ret == -1) return -1;

  return 0;
}

/*************************************************
Function:       cardCorner4
Description:    应用自适应二值化61,20分割图片，去除肤色和高光，定位角点
Calls:          cornerSearch：卡片轮廓多边形拟合，获取符合条件的卡片角点
Input:          card_gray: 卡片灰度图
                skinMask: 肤色掩膜
                card_w: 卡片宽度
                card_h: 卡片高度
Output:         corner: 定位到的角点
                areaRatio: 白点占比
Return:         正确提取4个角点返回0，否则返回-1
*************************************************/
int cardCorner4(Mat& card_gray, Mat& skinMask, vector<Point>& corner,
                float& areaRatio, int card_w, int card_h) {
  corner.clear();
  int ret = 0;
  Mat card_threshold1, card_threshold2, highlightMask;
  // highlightDet1(card_gray, highlightMask);
  threshold(card_gray, highlightMask, 247, 255, cv::THRESH_BINARY);

  adaptiveThreshold(card_gray, card_threshold1, 255, CV_ADAPTIVE_THRESH_MEAN_C,
                    CV_THRESH_BINARY, 61, 20);
  card_threshold2 = card_threshold1 - skinMask - highlightMask;
  // TODO
  // 修改点：调用cornerSearch方法之前仅做功能相关操作，将图像优化相关操作放在cornerSearch中统一完成
  ret = cornerSearch(card_gray, card_threshold2, corner, areaRatio, card_w,
                     card_h);
  if (ret == -1) return -1;
  return 0;
}

/*************************************************
Function:       cardCorner5
Description:    应用自适应二值化61,20分割图片，定位角点
Calls:          cornerSearch：卡片轮廓多边形拟合，获取符合条件的卡片角点
Input:          card_gray: 卡片灰度图
                skinMask: 肤色掩膜
                card_w: 卡片宽度
                card_h: 卡片高度
Output:         corner: 定位到的角点
                areaRatio: 白点占比
Return:         正确提取4个角点返回0，否则返回-1
*************************************************/
int cardCorner5(Mat& card_gray, Mat& skinMask, vector<Point>& corner,
                float& areaRatio, int card_w, int card_h) {
  corner.clear();
  int ret = 0;
  Mat card_threshold1, card_threshold2, highlightMask;
  adaptiveThreshold(card_gray, card_threshold1, 255, CV_ADAPTIVE_THRESH_MEAN_C,
                    CV_THRESH_BINARY, 61, 20);
  card_threshold2 = card_threshold1;  // - skinMask - highlightMask;
  // TODO
  // 修改点：调用cornerSearch方法之前仅做功能相关操作，将图像优化相关操作放在cornerSearch中统一完成
  ret = cornerSearch(card_gray, card_threshold2, corner, areaRatio, card_w,
                     card_h);
  if (ret == -1) return -1;
  return 0;
}

/*************************************************
Function:       cardCorner_highlight
Description:
应用两种阈值自适应二值化，并取去掉高光和保留高光效果最好的情况定位角点
Calls:   cornerSearch：   卡片轮廓多边形拟合，获取符合条件的卡片角点
Input:          card_gray: 卡片灰度图
                                skinMask: 肤色掩膜
                                card_w: 卡片宽度
                                card_h: 卡片高度
Output:         corner:  定位到的角点
                                areaRatio: 白点占比
Return: 正确提取四个角点返回0，否则返回-1
*************************************************/
int cardCorner_highlight(Mat& card_gray, Mat& skinMask, vector<Point>& corner,
                         float& areaRatio, int card_w, int card_h) {
  int ret = 0;
  Mat bright_exposure_binary, bright_exposure_sub_highlight,
      bright_exposure_sub_skin, bright_exposure_sub_highlight_skin,
      dark_exposure_binary, highlight_Mask;
  float areaRatio_tmp = 0.0, areaRatio_tmp1 = 0.0;
  vector<Point> corner_tmp, corner_tmp1;

  corner.clear();
  // TODO
  // 修改点1：流程修改。优化生成二值图的处理顺序，按照处理分支逐一进行图像处理操作,某个分支验证成功，其他的分支是不用执行的
  // 去肤色保留高光，对应曝光较暗的情况
  adaptiveThreshold(card_gray, dark_exposure_binary, 255,
                    CV_ADAPTIVE_THRESH_MEAN_C, CV_THRESH_BINARY, 453, -25);

  dark_exposure_binary = dark_exposure_binary - skinMask;
  // TODO
  // 修改点2：调用cornerSearch方法之前仅做功能相关操作，将图像优化相关操作放在cornerSearch中统一完成
  // 阈值2检测角点，白点占比大于0.88作为定位结果，否则继续
  ret = cornerSearch(card_gray, dark_exposure_binary, corner, areaRatio, card_w,
                     card_h);
  if (areaRatio > 0.88) return ret;
  // 检测高光区域
  // highlightDet1(card_gray, highlightMask);
  // TODO 修改点3：highlightDet函数重构
  threshold(card_gray, highlight_Mask, 247, 255, cv::THRESH_BINARY);
  // 两种阈值自适应二值化
  adaptiveThreshold(card_gray, bright_exposure_binary, 255,
                    CV_ADAPTIVE_THRESH_MEAN_C, CV_THRESH_BINARY, 453, -55);
  // 去高光，保留肤色
  bright_exposure_sub_highlight = bright_exposure_binary - highlight_Mask;
  // 保留肤色检测角点，白点占比大于0.87作为正确结果，否则继续
  //  cornerSearch 操作改变阈值图
  ret = cornerSearch(card_gray, bright_exposure_sub_highlight, corner,
                     areaRatio, card_w, card_h);
  // 去肤色检测角点，并比较去掉或保留高光的情况，效果最好的作为定位结果
  if ((ret == 0 && areaRatio < 0.87) || ret == -1) {
    // 高曝光图去肤色
    bright_exposure_sub_skin = bright_exposure_binary - skinMask;
    // 高曝光图去肤色、去高光区
    bright_exposure_sub_highlight_skin =
        bright_exposure_binary - skinMask - highlight_Mask;
    // 角点定位
    int ret2 = cornerSearch(card_gray, bright_exposure_sub_highlight_skin,
                            corner_tmp1, areaRatio_tmp1, card_w, card_h);
    int ret1 = cornerSearch(card_gray, bright_exposure_sub_skin, corner_tmp,
                            areaRatio_tmp, card_w, card_h);
    if (ret1 == -1 && ret2 == -1) return ret;
    // 取两种情况最优的结果与保留肤色情况作比较，最好的结果作为最终结果
    areaRatio_tmp =
        (areaRatio_tmp > areaRatio_tmp1) ? areaRatio_tmp : areaRatio_tmp1;
    corner_tmp = (areaRatio_tmp > areaRatio_tmp1) ? corner_tmp : corner_tmp1;
    if ((areaRatio_tmp > 0.82 && areaRatio_tmp > areaRatio)) {
      corner = corner_tmp;
      ret = 0;
    }
  }
  if (ret == -1) return -1;
  return 0;
}

/*************************************************
Function:       detect_card
Description:    卡片定位主函数
Calls:          cardCorner_highlight:
应用两种阈值自适应二值化，并取去掉高光和保留高光效果最好的情况定位角点
                                cardCorner2:
应用自适应二值化61,0分割图片，去除肤色，定位角点
                                cardCorner4:
应用自适应二值化61,20分割图片，去除肤色和高光，定位角点
                                cardCorner5:
应用自适应二值化61,20分割图片，保留肤色和高光，定位角点
Input:          src:
输入原图 camera: 新旧摄像头标识
Output:         card: 定位后卡片图像
                                isfront:是否为卡片正面
Return:         正确检测卡片返回0
*************************************************/
int get_card_rect(Mat& src,  cv::Mat & small_color, cv::Mat & small_gray, int card_w, int card_h,  int camera, vector<Point> &corner, Mat& card, bool& isfront, Rect& rectFace);
int detect_corner(Mat& src,  cv::Mat& small_color, cv::Mat &small_gray, int card_w, int card_h, int camera,  vector<Point> &corner);
int detect_card(Mat& src, Mat& card, bool& isfront, Rect& rectFace,
                int camera) {
  double start = tdr_get_current_time();

  /*新旧摄像头卡片尺寸*/
  int card_w, card_h;
  if (camera == 1) {  // 旧
    card_w = 247;     // 247
    card_h = 154;     // 154
  } else {            // 新
    card_w = 210;
    card_h = 132;
  }
  cv::Mat small_gray;
  cv::Mat small_color;
  int scale, width, height;
  scale = src.cols / 400;
  width = src.cols / scale;
  height = src.rows / scale;

  vector<Point> corner;
  int ret;

  // 卡片缩放到400x300
  resize(src, small_color, Size(width, height), cv::INTER_NEAREST);

  // 灰度化
  if (src.channels() > 1) {
    cvtColor(small_color, small_gray, CV_BGR2GRAY);
  } else {
    small_gray = small_color;
  }

#if 1
  //新的卡片AI检测算法
  std::vector<BoxfWithLandmarks> detected_boxes_kps;
  ret=detect_card_ai(small_color, detected_boxes_kps);
  if(ret!=0) return ret;

  //卡片AI后处理  
  ret=postprocess_card_ai(small_color, camera, detected_boxes_kps, corner, scale);
  if(ret!=0) return ret;
#else
  //旧的卡片检测算法
  if (camera == 1) {  // 旧
    card_w = 247;   card_h = 154; 
  } else {            // 新
    card_w = 210; card_h = 132;
  }
  ret=detect_corner(src, small_color, small_gray, card_w, card_h, camera,  corner);
  if(ret!=0) return ret;
#endif

  ret=get_card_rect(src, small_color, small_gray, card_w, card_h, camera,  corner, card, isfront, rectFace);
  if(ret!=0) return ret;
  double end = tdr_get_current_time();

#ifdef SUPPORT_TEST_CTL
  append_test_info_item("--卡片定位时间: ", (int)(end - start));
#endif

  LOGD("card detect: %.3f ms", end - start);

  return 0;

}



int detect_corner(Mat& src, Mat &small_color, cv::Mat &small_gray, int card_w, int card_h,  int camera,  vector<Point> &corner)
{
  int ret = 0;
  CardRect cardRect;
  corner.clear();

  // 肤色检测
  Mat skinMask = Mat::zeros(small_color.rows, small_color.cols, CV_8UC1);
  // TODO 修改点1：使用优化提速的SkinDetect7方法替换SkinDetect

//#ifndef CLOSE_SKIN
  if (SkinDetect7(small_color, skinMask) < 0) return -1;
//#endif

#if 0
  Mat skinMask_test = Mat::zeros(small_color.rows, small_color.cols, CV_8UC1);
  SkinDetect(small_color, skinMask_test);
  Mat diff_img;
  absdiff(skinMask, skinMask_test, diff_img);
  imshow("SkinDetect_diff_img", diff_img);
  waitKey(0);
#endif
  // 使用自适应二值化进行卡片粗定位。根据相机种类 分两种阈值
  Mat CardRegion_mask, noSkin_mask;
  adaptiveThreshold(small_gray, CardRegion_mask, 255, CV_ADAPTIVE_THRESH_MEAN_C,
                    CV_THRESH_BINARY, 201, (card_w > 230) ? 0 : 7);

  // 粗定位去肤色
  noSkin_mask = CardRegion_mask - skinMask;
  // 去噪，去除一些区域较小的孤立白色区域
  medianBlur(noSkin_mask, noSkin_mask, 3);
  Mat element = getStructuringElement(MORPH_RECT, Size(8, 8));
  morphologyEx(noSkin_mask, noSkin_mask, CV_MOP_OPEN, element);

  // 粗定位提取轮廓
  vector<vector<Point>> card_RoughContours;
  findContours(noSkin_mask, card_RoughContours, 0, 1);
  vector<vector<Point>> contours_poly(card_RoughContours.size());
  for (size_t i = 0; i < card_RoughContours.size(); i++) {
    // 获取轮廓外接矩形和最小矩形
    if (card_RoughContours[i].size() > 100) {
      approxPolyDP(Mat(card_RoughContours[i]), contours_poly[i], 10, true);
      Rect appRect(boundingRect(Mat(contours_poly[i])));
      cardRect.boundRect.push_back(appRect);
      cardRect.rotatedRect.push_back(minAreaRect(contours_poly[i]));
    }
  }

  if (cardRect.rotatedRect.size() == 0) return -3;
  // 获取面积最大的最小矩形作为粗定位结果
  int card_idx = 0;
  int area_max = 0;
  for (int i = 0; i < cardRect.rotatedRect.size(); i++) {
    int area_w = max(cardRect.rotatedRect[i].size.width,
                     cardRect.rotatedRect[i].size.height);
    int area_h = min(cardRect.rotatedRect[i].size.width,
                     cardRect.rotatedRect[i].size.height);
    int area_tmp = area_w * area_h;
    // 粗定位轮廓大小与卡片接近，直接作为结果
    if (abs(area_w - card_w) < 10 && abs(area_h - card_h) < 10) {
      card_idx = i;
      break;
    }
    // 获取最大面积的矩形区域
    if (area_tmp > area_max && area_w > card_h - 30 && area_h > card_h - 30) {
      area_max = area_tmp;
      card_idx = i;
    }
  }
  // 获取的区域扩大一部分，并保证在图片内
  Rect cardrect_ex;
  cardrect_ex.x = max(0, cardRect.boundRect[card_idx].x - 5);
  cardrect_ex.y = max(0, cardRect.boundRect[card_idx].y - 10);
  cardrect_ex.width = cardRect.boundRect[card_idx].width + 10;
  cardrect_ex.width = min(cardrect_ex.width, small_gray.cols - cardrect_ex.x - 1);
  cardrect_ex.height = cardRect.boundRect[card_idx].height + 15;
  cardrect_ex.height = min(cardrect_ex.height, small_gray.rows - cardrect_ex.y - 1);

  // 区域小于卡片为错误结果
  if (cardrect_ex.width < (card_h - 15) || cardrect_ex.height < (card_h - 15))
    return -3;

  // 根据粗定位结果，获取粗定位灰度图
  Mat card_gray = small_gray(cardrect_ex);
  // 获取粗定位图中的肤色区域
  Mat card_skinMask = skinMask(cardrect_ex);

  float areaRatio1 = 0.0;
  float areaRatio2 = 0.0;
  float areaRatio4 = 0.0;
  float areaRatio5 = 0.0;
  vector<Point> corner1, corner2, corner3, corner4, corner5;
  // 用不同阈值和方法定位卡片角点
  ret = cardCorner_highlight(card_gray, card_skinMask, corner1, areaRatio1,
                             card_w, card_h);  // ① 去高光mask
  if (ret == 0) {
    corner.assign(corner1.begin(), corner1.end());
  } else {
    ret = cardCorner2(card_gray, card_skinMask, corner2, areaRatio2, card_w,
                      card_h);  // ② 去肤色mask
    if (ret == 0) {
      corner.assign(corner2.begin(), corner2.end());
    } else {
      ret = cardCorner4(card_gray, card_skinMask, corner4, areaRatio4, card_w,
                        card_h);  // ③ 去肤色和高光mask
      if (ret == 0) {
        corner.assign(corner4.begin(), corner4.end());
      } else {
        ret = cardCorner5(card_gray, card_skinMask, corner5, areaRatio5, card_w,
                          card_h);  // ④ 换个二值化阈值再试试
        if (ret == 0) {
          corner.assign(corner5.begin(), corner5.end());
        }
      }
    }
  }

  if (ret == -1) return -3;

  // 保证角点位置不越界
  for (int i = 0; i < corner.size(); i++) {
    corner[i].x = max(0, corner[i].x + cardrect_ex.x);
    corner[i].x = min(corner[i].x, small_gray.cols - 1);
    corner[i].y = max(0, corner[i].y + cardrect_ex.y);
    corner[i].y = min(corner[i].y, small_gray.rows - 1);
  }

  return 0;
}



int get_card_rect(Mat& src, cv::Mat & small_color, cv::Mat &small_gray,  int card_w, int card_h, int camera,  vector<Point> &corner, Mat& card, bool& isfront, Rect& rectFace)
{
  int ret;
  float scale = src.cols / 400;
  // 使用找到的四个角点，拟合出一个旋转矩形roi_rect，即定位到的卡片区域
  RotatedRect roi_rect = minAreaRect(corner);

#ifdef TDR_TRACE
  cv::Mat showmat;
  if (isDebugShow()) {
        showmat= small_color.clone();
        cv::RNG rng(time(0));
        for(int p = 0; p < corner.size(); p++) {
            circle(showmat, corner[p] ,5,
            Scalar(rng.uniform(0,255),rng.uniform(0,255),rng.uniform(0,255)));
        }
        cv::Point2f vertex[4];
        roi_rect.points(vertex);
        for(int i=0; i<4; i++)
        {
              line(showmat, vertex[i] , vertex[(i+1)%4] ,
              Scalar(rng.uniform(0,255),rng.uniform(0,255),rng.uniform(0,255)));
        }
        cv::imshow("card_detect", showmat);
  }
#endif

// 有效位置判断，无效返回 -1002
// 判断左、右边界限制
#ifndef NO_JUDGE_CARD_POSITION 
#if 0
	sort(corner.begin(), corner.end(), compoint_x);
	if(isNewCamera()==1)  //wwwwwww
	{
//printf("corner0=%d, corner3=%d\n", corner[0].x,  corner[3].x);
			if(corner[0].x<20  || corner[3].x > 380) return ERROR_CARD_LOCATION;
			//if(corner[0].x<25  || corner[3].x > 370) return ERROR_CARD_LOCATION;
	}
	else{
//printf("corner0=%d, corner3=%d\n", corner[0].x,  corner[3].x);
			if(corner[0].x<8  || corner[3].x > 392) return ERROR_CARD_LOCATION;
			//if(corner[0].x<10  || corner[3].x > 385) return ERROR_CARD_LOCATION;
	}
#endif
  // 判断高度限制
  sort(corner.begin(), corner.end(), compoint_y);
  float line_locate_rate = (float)corner[0].y / (float)small_color.rows;
  //printf("line_locate_rate=%f; corner[0].y=%d\n", line_locate_rate,
  //       corner[0].y);
  //if (line_locate_rate < 0.3333) return ERROR_CARD_LOCATION;
  if (line_locate_rate < 0.3) return ERROR_CARD_LOCATION;
#endif

  ///////////////////////////////////
  int roi_rrect_w = max(roi_rect.size.width, roi_rect.size.height);
  int roi_rrect_h = min(roi_rect.size.width, roi_rect.size.height);

#ifndef USE_ALL_CARD
  //if (roi_rrect_w < (card_w - 20) || roi_rrect_w > (card_w + 20) ||
      //roi_rrect_h < (card_h - 20) || roi_rrect_h > (card_h + 20)) www
  if (roi_rrect_w < (card_w - 50) || roi_rrect_w > (card_w + 50) ||
      roi_rrect_h < (card_h - 50) || roi_rrect_h > (card_h + 50))
    return -3;
#endif

  /* 新摄像头角点位置扩大5个像素 */
  //if (camera != 1) {
    //roi_rect.size.width += 5;
    //roi_rect.size.height += 5;
  //}
  // 计算不越界的外接矩形
  // safeBoundRct主要是提取roi_rect内图像时在imgRotate_中使用，如果使用imgRotate_5则不需要此操作
  Rect_<float> safeBoundRct;
  bool isFormRect = calcSafeRect(roi_rect, small_gray, safeBoundRct);
  if (!isFormRect) {
    // LOGD("calcSafeRect ng");
    return -1;
  }

#ifdef TDR_TRACE
  if (isDebugShow()) {
    cv::rectangle(showmat, safeBoundRct, Scalar(0, 0, 255), 1, 8, 0);
    imshow("card_detect", showmat);
  }
#endif
  /*
  // safeBoundRct 对边界好像有扩展，不准，直接用点的坐标最低值更准确
  float line_locate_rate = (float)safeBoundRct.y / (float)small_color.rows;
  if (line_locate_rate < 0.3333) return ERROR_CARD_LOCATION;
  printf("safeBoundRct = %f, %f\n", safeBoundRct.x,
         safeBoundRct.x + safeBoundRct.width);

  if (isNewCamera() == 1)  // wwwwwww
  {
    if (safeBoundRct.x < 30 || safeBoundRct.x + safeBoundRct.width > 370)
      return ERROR_CARD_LOCATION;
  } else {
    if (safeBoundRct.x < 15 || safeBoundRct.x + safeBoundRct.width > 385)
      return ERROR_CARD_LOCATION;
  }
  */

  int scale1 = 1;
  Mat card_small;
  Mat faceTmp;
  Rect facerectTmp;
  // 在缩略图上根据四个角点的角度旋转，得到缩小图上的card图像card_small
  ret = imgRotate_2(small_color, card_small, safeBoundRct, roi_rect, scale1);
  if (ret == -1) return -1;

  isfront = false;
  // Mat cardTmp = cardFace.clone();

  // 检测人脸，并根据人脸位置翻转卡片。faceRet=2时，卡片倒置需要翻转
  int faceRet = getCardFaceImage(card_small, faceTmp, facerectTmp);
  if (faceRet == 1 || faceRet == 2) {
    isfront = true;
    rectFace.x = facerectTmp.x * scale;
    rectFace.y = facerectTmp.y * scale;
    rectFace.width = facerectTmp.width * scale;
    rectFace.height = facerectTmp.height * scale;
  }

  // 在原始图像(尺寸)上，提取出card的图像
  Mat srcbk = src.clone();
  if (isfront == false && camera != 1) {
#ifdef REMOVE_SIGNAL_LINE
    /*提取右侧白线起点区域*/
    // #ifdef REMOVE_SIGNAL_LINE
    Rect roi(1420, 300, 150, 300);
    Mat loc = srcbk(roi);
    Mat locimg = loc.clone();
    /*图像二值化和形态学处理*/
    cvtColor(locimg, locimg, cv::COLOR_BGR2GRAY);
    threshold(locimg, locimg, 0, 255, cv::THRESH_BINARY + cv::THRESH_OTSU);
    Mat element = getStructuringElement(cv::MORPH_RECT, Size(5, 5));
    morphologyEx(locimg, locimg, CV_MOP_OPEN, element);
    element = getStructuringElement(cv::MORPH_RECT, Size(1, 20));
    morphologyEx(locimg, locimg, CV_MOP_CLOSE, element);
    /*提取轮廓并根据大小定位白线起点*/
    vector<vector<Point>> lcontours;
    findContours(locimg, lcontours, 0, 1);
    Rect rslt;
    for (size_t i = 0; i < lcontours.size(); i++) {
      if (lcontours[i].size() > 100) {
        Rect appRect(boundingRect(lcontours[i]));
        if (appRect.width > 10 && appRect.height > 20 && appRect.height < 50) {
          rslt = appRect;
          break;
        }
      }
    }
    /*根据白线起点定位白线范围,中值滤波*/
    if (rslt.width > 0 && rslt.height > 0 && rslt.x > 0 && rslt.y > 0) {
      int yb1 = 300 + rslt.y + rslt.height / 5;
      int yb2 = 300 + rslt.y + rslt.height / 7 * 6;
      Rect lroi1(0, yb1, 1600, 10);
      Rect lroi(0, yb2, 1600, 8);
      Mat line1 = srcbk(lroi);
      Mat line2 = srcbk(lroi1);

      Mat l1 = line1.clone();
      Mat l2 = line2.clone();
      medianBlur(l1, l1, 5);
      medianBlur(l2, l2, 5);
      l1.copyTo(line1);
      l2.copyTo(line2);
    }
#endif
  }
  // TODO 待优化: 替换旋转API
  ret = imgRotate_2(srcbk, card, safeBoundRct, roi_rect, scale);
  if (ret == -1) return -1;
  if (isfront == true) {
    // 根据人脸的检测结果判断是否图片倒置，是需要转正
    if (faceRet == 2) flip(card, card, -1);
  }

  return ret;
}
