#include <omp.h>
#include <opencv2/core/core.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <vector>

#include "card_type.h"
#include "common.h"
#include "passport_visa.h"
#include "result.h"
#include "tdr_face_detect.h"
#include "tdrocr.h"

// const char * g_version_information = "tdrocr v3.0.5c-20210816";
const char* g_version_information = "tdrocr v3.5.0c-20231226";

void setTotalCount(int width, int height, int oldcamera);
void unload_model_centerface();
int getPassportFaceImage(cv::Mat& src, cv::Mat& faceimg);
int get_visa_issue_info(cv::Mat& src, OCR_RESULT& ocr_result);
int tdrocr_scan_passport(cv::Mat& src, OCR_RESULT& ocr_result);
void clearVisaIssueInfo();
int initVisaIssueInfo(char* path);
void uninit_crnn();
int isNewCamera();
int g_init_ok = 0;
extern int g_scan_none;
int init_card_ai(char * dir);
void uninit_card_ai();
extern cv::Mat g_faceimg, g_cropimg;

int g_scan_type = CERTIFICATE_TYPE_DEFAULT;
int tdr_passport_scan(Mat& src, OCR_RESULT& ocr_result);
int tdr_mixed_scan(Mat& src, OCR_RESULT& ocr_result);

std::string visa_type_mrz=  "ACDEFGLMRTUWXYZQPJXS"; //mrz中支持的签证代码
std::string visa_type_item1="WCDEFGLMRTUWXYZQJJXS"; 
std::string visa_type_item2="                21  ";
//Diplomatic 外交签证

void tdrocr_reset() {
  ocr_mean_conf = 0;
  ocr_min_conf = 0;
  pos_min_conf = -1;
  ocr_sum_conf = 0;
  clearResult();
}

int tdrocr_conf(int& sumConf, int& minConf, int& posMin) {
  minConf = ocr_min_conf;
  sumConf = ocr_sum_conf;
  posMin = pos_min_conf;
  return 0;
}

void tdrocr_unInit() {
  match_unInit();
  unload_model_centerface();
  uninit_crnn();
  g_init_ok = 0;
#ifdef SUPPORT_VISA
  clearVisaIssueInfo();
#endif

  uninit_card_ai();
  // 测试目录时需要清状态，
  // 主要是 g_get_issue_addr.im_valid = false; rotation中控制
  // 仿射变换逆矩阵的有效状态rotation
  // LOGI("clearVisaIssueInfo.");
}
int init_ncode_map();

int tdrocr_init(const char* dir, const char* opt, int isNewCamera) {
  int ret;
  std::string model_dir = dir;
  LOGI("version: %s", g_version_information);

  if (isNewCamera == 0) {
    LOGI(">>>> isOldCamera <<<<\n");
    setTotalCount(24, 34, 0);
  } else if (isNewCamera == 1) {
    // setTotalCount(21, 31, 0);
    LOGI(">>>> isNewCamera <<<<\n");
    setTotalCount(19, 28, 1);
  } else {
    LOGI("tdrocr_init: isNewCamera=%d is not support", isNewCamera);
    return -1;
  }
  init_ncode_map();
  int model_opt=0;
  if(strcasecmp(opt, "opencl")==0){
     model_opt=1;
  }
  ret = init_crnn((char *)dir, model_opt);
  if (ret != 0) {
    LOGI("===> init1 fail <===  %d", ret);
    return ret;
  }

  LOGI("init 1 pass.");

  ret = match_init(dir);

  if (ret != 0) {
    LOGI("===> init2 fail <===  %d", ret);
    return ret;
  }

  LOGI("init 2 pass.");

  ret = tdr_face_init2(model_dir.c_str());
  if (ret != 0) {
    LOGI("===> init3 fail <===  %d", ret);
    return ret;
  }

  LOGI("init 3 pass.");
#ifdef USE_ISSUER_ADDR
  ret = initVisaIssueInfo((char*)dir);
  if (ret != 0) {
    LOGI("===> init4 fail <===  %d", ret);
    return ret;
  }
  LOGI("init 4 pass.");
#endif

  ret=init_card_ai((char *)dir);
  if (ret != 0) {
    LOGI("===> init5 fail <===  %d", ret);
    return ret;
  }
  LOGI("init 5 pass.");


  LOGI(">>>> init success <<<<");

  memset((char *)&g_next_frame, 0x00, sizeof(g_next_frame));

  g_init_ok = 1;
  return 0;
}

int tdrocr_scan_yuv(unsigned char* src_data, int width, int height, int frm_wd,
                    int frm_hg, int mode, OCR_RESULT& ocr_result) {
  if (g_init_ok != 1) {
    clearOcrResult(ocr_result);
    LOGI("tdrocr_init first!");
    return -1;
  }
  int ret = 0;
  cv::Mat mBgr;
  // double start=tdr_get_current_time();
  cv::Mat src(height + height / 2, width, CV_8UC1, (unsigned char*)src_data);
  cv::cvtColor(src, mBgr, CV_YUV2BGR_NV21);
  ret = tdrocr_scan_mat(mBgr, mode, ocr_result);
  // double end=tdr_get_current_time();

  return ret;
}

int tdrocr_scan_mat(cv::Mat& src, int mode, OCR_RESULT& ocr_result) {
  clearOcrResult(ocr_result);
  if (g_init_ok != 1) {
    LOGI("tdrocr_init first!");
    LOGI("ocr return = %d", -1);
    return -1;
  }
  cv::Mat dst;
  if (src.cols != 1600) {
    cv::Mat img_resize;
    double scale = src.cols / 1600.0;
    int width = round(src.cols / scale);
    int height = round(src.rows / scale);
    cv::resize(src, img_resize, Size(width, height), cv::INTER_NEAREST);
    dst = img_resize;
  } else {
    dst = src;
  }

  if (dst.rows < 800) {
    LOGI("image size invalid.");
    LOGI("ocr return = %d", ERROR_IMAGE_SIZE);
    return ERROR_IMAGE_SIZE;
  }

  g_cropimg.release();
  g_faceimg.release();
  int ret = tdrocr_scan_passport(dst, ocr_result);
//if(!g_cropimg.empty()){
//imshow("g_cropimg", g_cropimg);
//cv::waitKey(0);
//}
  if(ret==0){
       g_next_frame.prev=0;
       g_detect_mode=0;
  }
  LOGI("ocr return = %d", ret);
  return ret;
}

int tdrocr_scan_passport(cv::Mat& src, OCR_RESULT& ocr_result) {
  if (g_init_ok != 1) {
    LOGI("tdrocr_init first!");
    return -1;
  }
#ifndef TDR_TRACE
  if (src.cols != 1600) {
    LOGI("image resolution %d X %d is not support!", src.cols, src.rows);
    return ERROR_IMAGE_SIZE;
  }
#endif
  int ret;
  switch (g_scan_type) {
    case CERTIFICATE_TYPE_PASSPORT:  // 1 护照
    case CERTIFICATE_TYPE_VISA:      // 7 签证
    case CERTIFICATE_TYPE_RESIDENT:  // 8 外国人居留许可
    case CERTIFICATE_TYPE_ENTRY_EXIT_PERMIT:
    case CERTIFICATE_TYPE_SEAFARER:
    case CERTIFICATE_TYPE_FROM_TW_ONEOFF:  // 4 台湾来内地
      LOGD("-----------CERTIFICATE_TYPE_PASSPORT/VISA/RESIDENT-----------\n");
      ret = tdr_passport_scan(src, ocr_result);
      break;
    case CERTIFICATE_TYPE_FROM_HM:  // 3 香港来内地
    case CERTIFICATE_TYPE_FROM_TW:  // 4 台湾来内地
    case CERTIFICATE_TYPE_APEC:
      LOGD("-----------CERTIFICATE_TYPE_FROM_HM/TW-----------\n");
      ret = tdr_card_scan(src, ocr_result);
      break;
    default:
      LOGD("-----------CERTIFICATE_TYPE_DEFAULT-----------\n");
      ret = tdr_mixed_scan(src, ocr_result);
      break;
  }

  return ret;
}

void tdrocr_set_scan_type(int scan_type) {
  switch (scan_type) {
    case CERTIFICATE_TYPE_DEFAULT:   // 0
    case CERTIFICATE_TYPE_PASSPORT:  // 1 护照
    // case CERTIFICATE_TYPE_ID:        // 2
    case CERTIFICATE_TYPE_FROM_HM:  // 3 香港来内地
    case CERTIFICATE_TYPE_FROM_TW:  // 4 台湾来内地
    case CERTIFICATE_TYPE_FROM_TW_ONEOFF:  // 4 台湾来内地

    // case CERTIFICATE_TYPE_TO_HM:     // 5
    // case CERTIFICATE_TYPE_TO_TW:     // 6
    case CERTIFICATE_TYPE_VISA:      // 7 签证
    case CERTIFICATE_TYPE_RESIDENT:  // 8 外国人居留许可
    case CERTIFICATE_TYPE_ENTRY_EXIT_PERMIT:
    case CERTIFICATE_TYPE_APEC:
    case CERTIFICATE_TYPE_SEAFARER:
      g_scan_type = scan_type;
      break;
    default:
      g_scan_type = CERTIFICATE_TYPE_DEFAULT;
      break;
  }

  LOGD("set_scan_type = %d", scan_type);
  g_detect_mode=0;

#ifdef SUPPORT_TEST_CTL
  append_test_info_item("tdrocr_set_scan_type: ", g_scan_type);
#endif
}

int tdr_passport_result_analysis(Mat& src, OCR_RESULT& ocr_result) {
  //double start = tdr_get_current_time();
  if(g_scan_none==1) return 0;

  bitwise_not(g_line12, g_line12_2);
  int ret = passportMatchOcr2(src, g_posRect, ocr_result);
  if (ret >= 0) {
// 识别成功后的处理,
// 签证页的附加识别，将识别到的信息存入ocr_result相应的字段
//      ocr_result.issue_at;   //签发地
//      ocr_result.add_id;     //右上角的证件号
/////////////////////////////////////////////////////////
#ifdef SUPPORT_VISA
    if (passport_ocr_result.result[0] == 'V' &&
        memcmp(passport_ocr_result.cntry1_result, "CHN", 3) == 0) {
      int rsp = get_visa_issue_info(src, ocr_result);


/////////签证种类校对，以机读区的结果为准 //////////
      char v=passport_ocr_result.result[1];
      if( strlen(ocr_result.category)<2){
          if(v!='<')
          {
              int npos=visa_type_mrz.find_first_of(v);
              if (npos != std::string::npos) {
                    ocr_result.category[0]= visa_type_item1[npos];
                    if(visa_type_item2[npos] != ' '){
                        ocr_result.category[1]=visa_type_item2[npos];
                        ocr_result.category[2]=0x00;
                    }
                    else{
                        ocr_result.category[1]=0x00;
                    }
              }
              else ocr_result.category[0]=0x00;
          }
          else ocr_result.category[0]=0x00;
      }
      else if( strlen(ocr_result.category)>=2){
          char v1=ocr_result.category[0];
          char v2=ocr_result.category[1];
          if(v=='J' && v1=='J' && (v2=='1' || v2=='2'))
          {
               ocr_result.category[2]=0x00;
          }
          else if(v!='<')
          {
              int npos=visa_type_mrz.find(v);
              if (npos != std::string::npos) {
                    ocr_result.category[0]= visa_type_item1[npos];
                    if(visa_type_item2[npos] != ' '){
                        ocr_result.category[1]=visa_type_item2[npos];
                        ocr_result.category[2]=0x00;
                    }
                    else if(visa_type_item2[npos] == ' ' && (v=='X' || v=='Q' || v=='S'))
                    {
                        if(v1==v && (v2=='1' || v2=='2'))
                        {
                              ocr_result.category[1]=v2;
                              ocr_result.category[2]=0x00;
                        }
                    }
                    else{
                        ocr_result.category[1]=0x00;
                    }
              }
              else ocr_result.category[0]=0x00;
          }
          else ocr_result.category[0]=0x00;
      }
///////////////////////////////////////////////////

      LOGD("-------------support visa------------\n");
    }
#else
#ifdef SUPPORT_TEST_CTL
    //append_test_info_item("不支持visa签发地址和签证号识别。", NO_TEST_PARA);
#endif
#endif
    /////////////////////////////////////////////////////////
    print_passport_result(ocr_result);

    cv::Mat faceimg;
    int count = getPassportFaceImage(src, faceimg);
    if (count == 0) {
      //ocr_result.faceimg = faceimg.clone();
      g_faceimg = faceimg.clone();
#ifdef TDR_TRACE
if(isDebugShow()) imshow("face", g_faceimg);
#endif
    }
  }

  cv::Mat cropimg;
  if (isNewCamera() == 1) {
    cropimg = src(Rect(120, 50, src.cols - 120 * 2, src.rows - 50 * 2));
  } else {
    cropimg = src(Rect(40, 40, src.cols - 40 * 2, src.rows - 40 * 2));
  }
  //ocr_result.cropimg = cropimg.clone();
  g_cropimg = cropimg.clone();

  //double end = tdr_get_current_time();
  //LOGD("tdr_passport_result_analysis cost %.3f ms", end - start);

  return ret;
}

int tdr_passport_scan(Mat& src, OCR_RESULT& ocr_result) {
#ifdef SUPPORT_TEST_CTL
  append_test_info_item("tdr_passport_scan: ", NO_TEST_PARA);
#endif
  double start = tdr_get_current_time();
  int sz1 = 0, sz2 = 0;
  int ret = passport_location(src, sz1, sz2);
  if (ret == 0) {
    ret = tdr_passport_result_analysis(src, ocr_result);
  }

  double end = tdr_get_current_time();
  LOGD("tdr_passport_scan: %.3f ms", end - start);

#ifdef SUPPORT_TEST_CTL
  append_test_info_item("tdr_passport_scan: ", (int)(end - start));
#endif
  return ret;
}

int tdr_mixed_scan(Mat& src, OCR_RESULT& ocr_result) {
#ifdef SUPPORT_TEST_CTL
  append_test_info_item("tdr_mixed_scan: ", NO_TEST_PARA);
#endif

  double start = tdr_get_current_time();
  int sz1 = 0, sz2 = 0;
  int ret = passport_location(src, sz1, sz2);
  if (ret < 0) {
    ret = tdr_card_scan(src, ocr_result);
  } else if (ret == 0) {
    ret = tdr_passport_result_analysis(src, ocr_result);
  } else {
    ret = -1;
    LOGD("error");
  }

  double end = tdr_get_current_time();
  LOGD("tdr_mixed_scan: %.3f ms", end - start);
#ifdef SUPPORT_TEST_CTL
  append_test_info_item("tdr_mixed_scan: ", (int)(end - start));
#endif
  return ret;
}
